# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import threading

import eventlet
from eventlet import greenpool
from keystoneclient import exceptions as k_exceptions
from keystoneclient.v2_0 import client as keyclient
from neutron_lib.db import model_base
from neutron_lib import exceptions as n_exc
from neutron_lib.plugins import constants as pconst
from neutron_lib import rpc as n_rpc
from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging
from oslo_serialization import jsonutils
from oslo_utils import excutils
import six
import sqlalchemy as sa

from gbpservice._i18n import _
from gbpservice.common import utils
from gbpservice.neutron.services.grouppolicy.common import constants as gconst
from gbpservice.neutron.services.servicechain.plugins.ncp import (
    exceptions as exc)
from gbpservice.neutron.services.servicechain.plugins.ncp import (
    model as ncp_model)
from gbpservice.neutron.services.servicechain.plugins.ncp import driver_base
from gbpservice.neutron.services.servicechain.plugins.ncp import plumber_base
from gbpservice.nfp.common import constants as nfp_constants
from gbpservice.nfp.common import topics as nfp_rpc_topics
from gbpservice.nfp.orchestrator.db import nfp_db as nfp_db


NFP_NODE_DRIVER_OPTS = [
    cfg.BoolOpt('is_service_admin_owned',
                help=_("Parameter to indicate whether the Service VM has to "
                       "be owned by the Admin"),
                default=False),
    cfg.IntOpt('service_create_timeout',
               default=nfp_constants.SERVICE_CREATE_TIMEOUT,
               help=_("Seconds to wait for service creation "
                      "to complete")),
    cfg.IntOpt('service_delete_timeout',
               default=nfp_constants.SERVICE_DELETE_TIMEOUT,
               help=_("Seconds to wait for service deletion "
                      "to complete")),
]
# REVISIT(ashu): Can we use is_service_admin_owned config from RMD
cfg.CONF.register_opts(NFP_NODE_DRIVER_OPTS, "nfp_node_driver")


LOG = logging.getLogger(__name__)

# REVISIT: L2 insertion not supported
GATEWAY_PLUMBER_TYPE = [pconst.FIREWALL, pconst.VPN]
nfp_context_store = threading.local()


class InvalidServiceType(exc.NodeCompositionPluginBadRequest):
    message = _("The NFP Node driver only supports the services "
                "VPN, Firewall and LB in a Service Chain")


class ServiceProfileRequired(exc.NodeCompositionPluginBadRequest):
    message = _("A Service profile is required in Service node")


class NodeVendorMismatch(exc.NodeCompositionPluginBadRequest):
    message = _("The NFP Node driver only handles nodes which have service "
                "profile with vendor name %(vendor)s")


class DuplicateServiceTypeInChain(exc.NodeCompositionPluginBadRequest):
    message = _("The NFP Node driver does not support duplicate "
                "service types in same chain")


class RequiredProfileAttributesNotSet(exc.NodeCompositionPluginBadRequest):
    message = _("The required attributes in service profile are not present")


class InvalidNodeOrderInChain(exc.NodeCompositionPluginBadRequest):
    message = _("The NFP Node driver does not support the order "
                "of nodes defined in the current service chain spec, "
                "order should be : %(node_order)s")


class UnSupportedServiceProfile(exc.NodeCompositionPluginBadRequest):
    message = _("The NFP Node driver does not support this service "
                "profile with service type %(service_type)s and vendor "
                "%(vendor)s")


class UnSupportedInsertionMode(exc.NodeCompositionPluginBadRequest):
    message = _("The NFP Node driver supports only L3 Insertion "
                "mode")


class ServiceInfoNotAvailableOnUpdate(n_exc.NeutronException):
    message = _("Service information is not available with Service Manager "
                "on node update")


class VipNspNotSetonProvider(n_exc.NeutronException):
    message = _("Network Service policy for VIP IP address is not configured "
                "on the Providing Group")


class NodeInstanceDeleteFailed(n_exc.NeutronException):
    message = _("Node instance delete failed in NFP Node driver")


class NodeInstanceCreateFailed(n_exc.NeutronException):
    message = _("Node instance create failed in NFP Node driver")


class NodeInstanceUpdateFailed(n_exc.NeutronException):
    message = _("Node instance update failed in NFP Node driver")


class OperationNotSupported(exc.NodeCompositionPluginBadRequest):
    message = _("The NFP Node driver doesn't support operation, "
                "if instance status is in BUILD state.")


class ServiceNodeInstanceNetworkFunctionMapping(model_base.BASEV2):
    """ServiceChainInstance to NFP network function mapping."""

    __tablename__ = 'ncp_node_instance_network_function_mappings'
    sc_instance_id = sa.Column(sa.String(36),
                               nullable=False, primary_key=True)
    sc_node_id = sa.Column(sa.String(36),
                           nullable=False, primary_key=True)
    network_function_id = sa.Column(sa.String(36),
                                    nullable=True)
    status = sa.Column(sa.String(20), nullable=True)
    status_details = sa.Column(sa.String(4096), nullable=True)


class NFPClientApi(object):
    """ Client side of the NFP Framework user """

    RPC_API_VERSION = '1.0'

    def __init__(self, topic):
        target = oslo_messaging.Target(
            topic=topic, version=self.RPC_API_VERSION)
        self.client = n_rpc.get_client(target)

    def create_network_function(self, context, network_function):
        LOG.info("Sending RPC CREATE NETWORK FUNCTION to Service "
                 "Orchestrator for tenant:%(tenant_id)s with "
                 "service profile:%(service_profile_id)s",
                 {'tenant_id': network_function['tenant_id'],
                  'service_profile_id': network_function[
                                            'service_profile']['id']})
        cctxt = self.client.prepare(
            fanout=False, topic=nfp_rpc_topics.NFP_NSO_TOPIC)
        return cctxt.cast(
            context,
            'create_network_function',
            network_function=network_function)

    def delete_network_function(self, context, network_function_id,
                                network_function_data):
        LOG.info("Sending RPC DELETE NETWORK FUNCTION to Service "
                 "Orchestrator for NF:"
                 "%(network_function_id)s",
                 {'network_function_id': network_function_id})
        cctxt = self.client.prepare(version=self.RPC_API_VERSION)
        return cctxt.cast(
            context,
            'delete_network_function',
            network_function_id=network_function_id,
            network_function_data=network_function_data)

    def update_network_function(self, context, network_function_id, config):
        LOG.info("Sending RPC UPDATE NETWORK FUNCTION to Service "
                 "Orchestrator for NF:"
                 "%(network_function_id)s",
                 {'network_function_id': network_function_id})
        cctxt = self.client.prepare(version=self.RPC_API_VERSION)
        return cctxt.cast(
            context,
            'update_network_function',
            network_function_id=network_function_id,
            config=config)

    def get_network_function(self, context, network_function_id):
        LOG.debug("Sending RPC GET NETWORK FUNCTION to Service "
                  "Orchestrator for NF: %s", network_function_id)
        cctxt = self.client.prepare(version=self.RPC_API_VERSION)
        return cctxt.call(
            context,
            'get_network_function',
            network_function_id=network_function_id)

    def consumer_ptg_added_notification(self, context, network_function_id,
                                        policy_target_group):
        LOG.info("Sending RPC CONSUMER PTG ADDED NOTIFICATION to Service "
                 "Orchestrator for NF:"
                 "%(network_function_id)s",
                 {'network_function_id': network_function_id})
        cctxt = self.client.prepare(version=self.RPC_API_VERSION)
        return cctxt.cast(context,
                   'consumer_ptg_added_notification',
                   network_function_id=network_function_id,
                   policy_target_group=policy_target_group)

    def consumer_ptg_removed_notification(self, context, network_function_id,
                                          policy_target_group):
        LOG.info("Sending RPC CONSUMER PTG REMOVED NOTIFICATION to "
                 " Service Orchestrator for NF:%(network_function_id)s",
                 {'network_function_id': network_function_id})
        cctxt = self.client.prepare(version=self.RPC_API_VERSION)
        return cctxt.cast(context,
                   'consumer_ptg_removed_notification',
                   network_function_id=network_function_id,
                   policy_target_group=policy_target_group)

    def policy_target_added_notification(self, context, network_function_id,
                                         policy_target):
        LOG.info("Sending RPC POLICY TARGET ADDED NOTIFICATION to "
                 "Service Orchestrator for NF:%(network_function_id)s",
                 {'network_function_id': network_function_id})
        cctxt = self.client.prepare(version=self.RPC_API_VERSION)
        return cctxt.cast(context,
                   'policy_target_added_notification',
                   network_function_id=network_function_id,
                   policy_target=policy_target)

    def policy_target_removed_notification(self, context, network_function_id,
                                           policy_target):
        LOG.info("Sending RPC POLICY TARGET REMOVED NOTIFICATION to "
                 "Service Orchestrator for NF:%(network_function_id)s",
                 {'network_function_id': network_function_id})
        cctxt = self.client.prepare(version=self.RPC_API_VERSION)
        return cctxt.cast(context,
                   'policy_target_removed_notification',
                   network_function_id=network_function_id,
                   policy_target=policy_target)

    def get_plumbing_info(self, context, node_driver_ctxt):
        LOG.info("Sending RPC GET PLUMBING INFO to Service Orchestrator ")
        request_info = dict(profile=node_driver_ctxt.current_profile,
                            tenant_id=node_driver_ctxt.provider['tenant_id'],
                            provider=node_driver_ctxt.provider)
        cctxt = self.client.prepare(version=self.RPC_API_VERSION)
        return cctxt.call(context, 'get_plumbing_info',
                          request_info=request_info)


class NFPContext(object):

    @staticmethod
    def store_nfp_context(sc_instance_id, **context):
        if not hasattr(nfp_context_store, 'context'):
            nfp_context_store.context = {}

        # Considering each store request comes with one entry
        if not nfp_context_store.context.get(sc_instance_id):
            NFPContext._initialise_attr(sc_instance_id)
        nfp_context_store.context[sc_instance_id].update(context)

    @staticmethod
    def clear_nfp_context(sc_instance_id):
        if not hasattr(nfp_context_store, 'context'):
            return
        if nfp_context_store.context.get(sc_instance_id):
            del nfp_context_store.context[sc_instance_id]

    @staticmethod
    def get_nfp_context(sc_instance_id):
        if not hasattr(nfp_context_store, 'context'):
            return {}
        if nfp_context_store.context.get(sc_instance_id):
            return nfp_context_store.context[sc_instance_id]
        return {}

    @staticmethod
    def _initialise_attr(sc_instance_id):
        context = {'thread_pool': greenpool.GreenPool(10),
                   'active_threads': [],
                   'sc_node_count': 0,
                   'sc_gateway_type_nodes': [],
                   'network_functions': [],
                   'update': False}
        if nfp_context_store.context:
            nfp_context_store.context.update({sc_instance_id: context})
        else:
            nfp_context_store.context = {sc_instance_id: context}


class NFPNodeDriver(driver_base.NodeDriverBase):
    SUPPORTED_SERVICE_TYPES = [
        pconst.FIREWALL, pconst.VPN,
        pconst.LOADBALANCERV2]
    vendor_name = nfp_constants.NFP_VENDOR.upper()
    required_heat_resources = {
        pconst.LOADBALANCERV2: ['OS::Neutron::LBaaS::LoadBalancer',
                                'OS::Neutron::LBaaS::Listener',
                                'OS::Neutron::LBaaS::Pool'],
        pconst.FIREWALL: ['OS::Neutron::Firewall',
                          'OS::Neutron::FirewallPolicy'],
        pconst.VPN: ['OS::Neutron::VPNService'],
    }
    initialized = False

    def __init__(self):
        super(NFPNodeDriver, self).__init__()
        self._lbaas_plugin = None
        self.nfp_db = nfp_db.NFPDbBase()

    @property
    def name(self):
        return self._name

    def initialize(self, name):
        self.initialized = True
        self._name = name
        if cfg.CONF.nfp_node_driver.is_service_admin_owned:
            self.resource_owner_tenant_id = self._resource_owner_tenant_id()
        else:
            self.resource_owner_tenant_id = None
        self._setup_rpc()

    def _setup_rpc(self):
        self.nfp_notifier = NFPClientApi(nfp_rpc_topics.NFP_NSO_TOPIC)

    def _parse_service_flavor_string(self, service_flavor_str):
        service_details = {}
        if ',' not in service_flavor_str:
            service_details['device_type'] = 'nova'
            service_details['service_vendor'] = service_flavor_str
        else:
            service_flavor_dict = dict(item.split('=') for item
                                       in service_flavor_str.split(','))
            service_details = {key.strip(): value.strip() for key, value
                               in six.iteritems(service_flavor_dict)}
        return service_details

    def _update_node_instance_network_function_map(self, context,
                                    updated_network_function_map):
        self.nfp_db.update_node_instance_network_function_map(
            context.plugin_session,
            context.current_node['id'],
            context.instance['id'],
            updated_network_function_map)

    def get_plumbing_info(self, context):
        plumbing_request = {}
        context._plugin_context = self._get_resource_owner_context(
            context._plugin_context)
        service_type = context.current_profile['service_type']

        service_flavor_str = context.current_profile['service_flavor']
        service_details = self._parse_service_flavor_string(service_flavor_str)
        if service_details['device_type'] == 'None':
            if not NFPContext.get_nfp_context(context.instance['id']):
                nfp_context = NFPContext.store_nfp_context(
                                context.instance['id'])
            return {}
        # Management PTs are managed by NFP since it supports hosting multiple
        # logical services in a single device
        # plumber will return stitching network PT instead of consumer
        # as chain is instantiated while creating provider group.
        if service_type in GATEWAY_PLUMBER_TYPE:
            gateway_type_node = {'service_type': service_type,
                                 'context': {}}
            nfp_context = NFPContext.get_nfp_context(context.instance['id'])
            if nfp_context:
                if len(nfp_context['sc_gateway_type_nodes']):
                    LOG.info(
                        "Not requesting plumber for PTs for service type "
                        "%(service_type)s", {'service_type': service_type})
                    if not nfp_context['update']:
                        nfp_context['sc_gateway_type_nodes'].append(
                            gateway_type_node)
                        NFPContext.store_nfp_context(
                            context.instance['id'],
                            sc_gateway_type_nodes=(
                                nfp_context['sc_gateway_type_nodes']))
                        return {}
                if not nfp_context['update']:
                    nfp_context['sc_gateway_type_nodes'].append(
                        gateway_type_node)
                    NFPContext.store_nfp_context(
                        context.instance['id'],
                        sc_gateway_type_nodes=(
                            nfp_context['sc_gateway_type_nodes']))
                    plumbing_request = self.nfp_notifier.get_plumbing_info(
                            context._plugin_context, context)
            else:
                NFPContext.store_nfp_context(
                    context.instance['id'],
                    sc_gateway_type_nodes=[gateway_type_node])
                plumbing_request = self.nfp_notifier.get_plumbing_info(
                        context._plugin_context, context)

        else:  # Loadbalancer which is one arm
            NFPContext.store_nfp_context(
                context.instance['id'])
            plumbing_request = self.nfp_notifier.get_plumbing_info(
                    context._plugin_context, context)

        LOG.info("Requesting plumber for PTs for "
                 "service type %(service_type)s with "
                 "%(plumbing_request)s ",
                 {'plumbing_request': plumbing_request,
                  'service_type': service_type})
        return plumbing_request

    def validate_create(self, context):
        if not context.current_profile:
            raise ServiceProfileRequired()
        if (not context.current_profile['vendor'] or not
            context.current_profile['insertion_mode'] or not
            context.current_profile['service_type'] or not
            context.current_profile['service_flavor']):
            raise RequiredProfileAttributesNotSet()
        if context.current_profile['vendor'] != self.vendor_name:
            raise NodeVendorMismatch(vendor=self.vendor_name)
        if (context.current_profile['insertion_mode'].lower() !=
                nfp_constants.L3_INSERTION_MODE):
            raise UnSupportedInsertionMode()
        if context.current_profile['service_type'] not in (
            self.SUPPORTED_SERVICE_TYPES):
            raise InvalidServiceType()
        self._is_node_order_in_spec_supported(context)

    def _validate_status(self, context):
        servicechain_instance_id = context.instance['id']
        servicechain_instance = context.sc_plugin.get_servicechain_instance(
                context.plugin_context, servicechain_instance_id)
        if servicechain_instance['status']:
            if servicechain_instance['status'] == nfp_constants.BUILD:
                raise OperationNotSupported()

    def validate_update(self, context):
        NFPContext.store_nfp_context(context.instance['id'],
                                     update=True)
        if not context.original_node:  # PT create/delete notifications
            return
        if context.current_node and not context.current_profile:
            raise ServiceProfileRequired()
        if context.current_profile['vendor'] != self.vendor_name:
            raise NodeVendorMismatch(vendor=self.vendor_name)
        if (context.current_profile['insertion_mode'].lower() !=
                nfp_constants.L3_INSERTION_MODE):
            raise UnSupportedInsertionMode()
        if context.current_profile['service_type'] not in (
            self.SUPPORTED_SERVICE_TYPES):
            raise InvalidServiceType()
        self._validate_status(context)

    def _wait(self, thread, context):
        try:
            result = thread.wait()
            return result
        except Exception as e:
            NFPContext.clear_nfp_context(context.instance['id'])
            raise e

    def create(self, context):
        try:
            context._plugin_context = self._get_resource_owner_context(
                context._plugin_context)
            network_function_id = None
            status = nfp_constants.BUILD
            status_details = 'processing create in node driver.'
            self._create_network_function(context)
        except Exception:
            # NFPContext.clear_nfp_context(context.instance['id'])
            exc_type, exc_value, exc_traceback = sys.exc_info()
            message = "Traceback: %s" % (exc_value)
            LOG.error(message)

        finally:
            self.nfp_db.create_node_instance_network_function_map(
                context.plugin_session, context.current_node['id'],
                context.instance['id'], network_function_id,
                status, status_details)
            LOG.info("Processed create NF in node driver."
                    "servicechain_instance_id: %(sci_id)s, "
                    "servicechain_node_id: %(scn_id)s", {
                        'sci_id': context.instance['id'],
                        'scn_id': context.current_node['id']})

    def _wait_for_node_operation_completion(self, context, network_function_id,
                                            operation):
        # Check for NF status in a separate thread
        LOG.debug("Spawning thread for nf ACTIVE poll operation: %s", (
            operation))
        nfp_context = NFPContext.get_nfp_context(context.instance['id'])
        nfp_context['sc_node_count'] -= 1
        nfp_context['network_functions'].append(network_function_id)
        # At last wait for the threads to complete, success/failure/timeout
        if nfp_context['sc_node_count'] == 0:
            network_functions = nfp_context['network_functions']
            for network_function in network_functions:
                LOG.debug("Spawning thread for nf ACTIVE poll")
                if operation == nfp_constants.DELETE:
                    gth = nfp_context['thread_pool'].spawn(
                        self._wait_for_network_function_delete_completion,
                        context, network_function)
                else:
                    gth = nfp_context['thread_pool'].spawn(
                        self._wait_for_network_function_operation_completion,
                        context, network_function, operation=operation)

                nfp_context['active_threads'].append(gth)

            message = "Active Threads count (%d), sc_node_count (%d)" % (
                len(nfp_context['active_threads']),
                nfp_context['sc_node_count'])
            LOG.debug(message)
            nfp_context['thread_pool'].waitall()
            # Get the results
            for gth in nfp_context['active_threads']:
                self._wait(gth, context)

            NFPContext.clear_nfp_context(context.instance['id'])
        else:
            NFPContext.store_nfp_context(context.instance['id'], **nfp_context)

    def update(self, context):
        NFPContext.clear_nfp_context(context.instance['id'])
        context._plugin_context = self._get_resource_owner_context(
            context._plugin_context)
        network_function_map = (
            self.nfp_db.get_node_instance_network_function_map(
                context.plugin_session,
                context.current_node['id'],
                context.instance['id']))

        if not all([network_function_map, context.original_node.get('config'),
                    context.current_node.get('config')]):
            return

        network_function_id = network_function_map.network_function_id
        self._update(context, network_function_id)

    def _get_node_count(self, context):
        current_specs = context.relevant_specs
        node_list = []
        for spec in current_specs:
            node_list.extend(spec['nodes'])
        NFPContext.store_nfp_context(context.instance['id'],
                                     sc_node_count=len(node_list))
        return len(node_list)

    def get_status(self, context):
        context._plugin_context = self._get_resource_owner_context(
            context._plugin_context)
        network_function_map = (
            self.nfp_db.get_node_instance_network_function_map(
                context.plugin_session,
                context.current_node['id'],
                context.instance['id']))
        nf_status = network_function_map.status
        if nf_status not in nfp_constants.NFP_STATUS:
            nf_status = nfp_constants.BUILD
        status = nfp_constants.NFP_STATUS_MAP[nf_status]
        return status

    def delete(self, context):
        nfp_context = (
            NFPContext.get_nfp_context(context.instance['id']))
        if nfp_context and not nfp_context.get('sc_node_count'):
            nfp_context['sc_node_count'] = self._get_node_count(context)

        context._plugin_context = self._get_resource_owner_context(
            context._plugin_context)
        network_function_map = (
            self.nfp_db.get_node_instance_network_function_map(
                context.plugin_session,
                context.current_node['id'],
                context.instance['id']))
        network_function_id = None
        if network_function_map:
            updated_network_function_map = {
                'status': nfp_constants.BUILD,
                'status_details': 'node driver processing node delete'}
            self._update_node_instance_network_function_map(
                context, updated_network_function_map)
            network_function_id = network_function_map.network_function_id

        if network_function_id:
            try:
                self._delete_network_function(context, network_function_id)
            except Exception:
                # NFPContext.clear_nfp_context(context.instance['id'])
                LOG.exception("Delete Network service Failed")
                exc_type, exc_value, exc_traceback = sys.exc_info()
                message = "Traceback: %s" % (exc_value)
                LOG.error(message)

        self._update_ptg(context)

    def update_policy_target_added(self, context, policy_target):
        if context.current_profile['service_type'] == pconst.LOADBALANCERV2:
            if self._is_service_target(policy_target):
                return
            context._plugin_context = self._get_resource_owner_context(
                context._plugin_context)
            network_function_map = (
                self.nfp_db.get_node_instance_network_function_map(
                    context.plugin_session,
                    context.current_node['id'],
                    context.instance['id']))
            if network_function_map:
                updated_network_function_map = {
                    'status': nfp_constants.BUILD,
                    'status_details': 'node driver processing PT add'}
                self._update_node_instance_network_function_map(
                    context, updated_network_function_map)
                network_function_id = network_function_map.network_function_id
                self.nfp_notifier.policy_target_added_notification(
                    context.plugin_context, network_function_id, policy_target)

    def update_policy_target_removed(self, context, policy_target):
        if context.current_profile['service_type'] == pconst.LOADBALANCERV2:
            if self._is_service_target(policy_target):
                return
            context._plugin_context = self._get_resource_owner_context(
                context._plugin_context)
            network_function_map = (
                self.nfp_db.get_node_instance_network_function_map(
                    context.plugin_session,
                    context.current_node['id'],
                    context.instance['id']))

            if network_function_map:
                updated_network_function_map = {
                    'status': nfp_constants.BUILD,
                    'status_details': 'node driver processing PT remove'}
                self._update_node_instance_network_function_map(
                    context, updated_network_function_map)
                network_function_id = network_function_map.network_function_id
                self.nfp_notifier.policy_target_removed_notification(
                    context.plugin_context, network_function_id, policy_target)

    def notify_chain_parameters_updated(self, context):
        pass  # We are not using the classifier specified in redirect Rule

    def update_node_consumer_ptg_added(self, context, policy_target_group):

        # When a group is created which is both consumer and provider.
        # method is invoked for stitching group too.. ignoring.
        if policy_target_group.get('proxied_group_id'):
            return
        if context.current_profile['service_type'] == pconst.FIREWALL:
            context._plugin_context = self._get_resource_owner_context(
                context._plugin_context)
            network_function_map = (
                self.nfp_db.get_node_instance_network_function_map(
                    context.plugin_session,
                    context.current_node['id'],
                    context.instance['id']))

            if network_function_map:
                updated_network_function_map = {
                    'status': nfp_constants.BUILD,
                    'status_details': 'node driver processing consumer add'}
                self._update_node_instance_network_function_map(context,
                        updated_network_function_map)
                network_function_id = network_function_map.network_function_id
                self.nfp_notifier.consumer_ptg_added_notification(
                    context.plugin_context, network_function_id,
                    policy_target_group)

    def update_node_consumer_ptg_removed(self, context, policy_target_group):
        # When a group is created which is both consumer and provider.
        # method is invoked for stitching group too.. ignoring.
        if policy_target_group.get('proxied_group_id'):
            return
        if context.current_profile['service_type'] == pconst.FIREWALL:
            context._plugin_context = self._get_resource_owner_context(
                context._plugin_context)
            network_function_map = (
                self.nfp_db.get_node_instance_network_function_map(
                    context.plugin_session,
                    context.current_node['id'],
                    context.instance['id']))

            if network_function_map:
                updated_network_function_map = {
                    'status': nfp_constants.BUILD,
                    'status_details': 'node driver processing consumer remove'}
                self._update_node_instance_network_function_map(
                    context, updated_network_function_map)
                network_function_id = network_function_map.network_function_id
                self.nfp_notifier.consumer_ptg_removed_notification(
                    context.plugin_context, network_function_id,
                    policy_target_group)

    def policy_target_group_updated(self, context, old_ptg, current_ptg):
        if not (old_ptg and current_ptg):
            return
        if current_ptg['description']:
            desc = current_ptg['description'].split(':')
            if 'opflex_eoc' in desc:
                if (set(old_ptg[
                        'provided_policy_rule_sets']).symmetric_difference(
                            set(current_ptg['provided_policy_rule_sets']))):
                    pts = context.gbp_plugin.get_policy_targets(
                            context.plugin_context,
                            filters={'port_id': [desc[-1]]})
                    (pt,) = pts
                    filters = {'description': [current_ptg['description']]}
                    ptgs = context.gbp_plugin.get_policy_target_groups(
                                context.plugin_context, filters)
                    prs = []
                    for ptg in ptgs:
                        prs += ptg['provided_policy_rule_sets']
                    context.gbp_plugin.update_policy_target_group(
                            context.plugin_context,
                            pt['policy_target_group_id'],
                            {'policy_target_group':
                                {'provided_policy_rule_sets':
                                    dict((x, '') for x in prs)}})

    def _wait_for_network_function_delete_completion(self, context,
                                                     network_function_id):
        # [REVISIT: (akash) do we need to do error handling here]
        if not network_function_id:
            return

        time_waited = 0
        network_function = None
        while time_waited < cfg.CONF.nfp_node_driver.service_delete_timeout:
            network_function = self.nfp_notifier.get_network_function(
                context.plugin_context, network_function_id)
            if not network_function or (
                    network_function['status'] == nfp_constants.ERROR):
                break
            eventlet.sleep(5)
            time_waited = time_waited + 5

        if network_function:
            LOG.error("Delete network function %(network_function)s "
                      "failed",
                      {'network_function': network_function_id})
            raise NodeInstanceDeleteFailed()

    def _wait_for_network_function_operation_completion(self, context,
                                                        network_function_id,
                                                        operation):
        if not network_function_id:
            raise NodeInstanceCreateFailed()

        time_waited = 0
        network_function = None
        timeout = cfg.CONF.nfp_node_driver.service_create_timeout

        while time_waited < timeout:
            network_function = self.nfp_notifier.get_network_function(
                context.plugin_context, network_function_id)
            LOG.debug("Got %s nf result for NF: %s with status:%s,"
                      "time waited: %s", (network_function_id, operation,
                      time_waited, network_function['status']))
            if not network_function:
                LOG.error("Failed to retrieve network function")
                eventlet.sleep(5)
                time_waited = time_waited + 5
                continue
            else:
                if time_waited == 0:
                    LOG.info("STARTED POLLING for %(operation)s network "
                             "function for NF:%(network_function_id)s "
                             "with initial result: %(result)s ",
                             {'operation': operation,
                              'network_function_id': network_function_id,
                              'result': network_function})
            if (network_function['status'] == nfp_constants.ACTIVE or
                network_function['status'] == nfp_constants.ERROR):
                LOG.info("COMPLETED POLLING for  %(operation)s network "
                         "function for NF:%(network_function_id)s ",
                         {'network_function_id': network_function_id,
                          'operation': operation})
                break
            eventlet.sleep(5)
            time_waited = time_waited + 5

        LOG.info("Got %(operation)s network function result for NF:"
                 "%(network_function_id)s with status:%(status)s",
                 {'network_function_id': network_function_id,
                  'operation': operation,
                  'status': network_function['status']})

        if network_function['status'] != nfp_constants.ACTIVE:
            LOG.error("%(operation)s network function:"
                      "%(network_function)s "
                      "failed. Status: %(status)s",
                      {'network_function': network_function_id,
                       'status': network_function['status'],
                       'operation': operation})
            if operation.lower() == nfp_constants.CREATE:
                raise NodeInstanceCreateFailed()
            elif operation.lower() == nfp_constants.UPDATE:
                raise NodeInstanceUpdateFailed()

    def _is_service_target(self, policy_target):
        if policy_target['name'] and (policy_target['name'].startswith(
            plumber_base.SERVICE_TARGET_NAME_PREFIX) or
            policy_target['name'].startswith('tscp_endpoint_service') or
            policy_target['name'].startswith('vip_pt')):
            return True
        else:
            return False

    def _resource_owner_tenant_id(self):
        user, pwd, tenant, auth_url = utils.get_keystone_creds()
        keystoneclient = keyclient.Client(username=user, password=pwd,
                                          auth_url=auth_url)
        try:
            tenant = keystoneclient.tenants.find(name=tenant)
            return tenant.id
        except k_exceptions.NotFound:
            with excutils.save_and_reraise_exception(reraise=True):
                LOG.error('No tenant with name %(tenant)s exists.',
                          {'tenant': tenant})
        except k_exceptions.NoUniqueMatch:
            with excutils.save_and_reraise_exception(reraise=True):
                LOG.error('Multiple tenants matches found for %(tenant)s',
                          {'tenant': tenant})

    def _get_resource_owner_context(self, plugin_context):
        # REVISIT(AKASH) Need to revisit as this api is not needed
        # with present scenarios
        '''
        if cfg.CONF.nfp_node_driver.is_service_admin_owned:
            resource_owner_context = plugin_context.elevated()
            resource_owner_context.tenant_id = self.resource_owner_tenant_id
            user, pwd, ignore_tenant, auth_url = utils.get_keystone_creds()
            keystoneclient = keyclient.Client(username=user, password=pwd,
                                              auth_url=auth_url)
            resource_owner_context.auth_token = keystoneclient.get_token(
                self.resource_owner_tenant_id)
            return resource_owner_context
        else:
            return plugin_context
        '''
        return plugin_context

    def _update(self, context, network_function_id):
        if (context.original_node['config'] != context.current_node['config']):
            try:
                updated_network_function_map = {
                    'status': nfp_constants.BUILD,
                    'status_details': 'node driver processing node update'}
                self._update_node_instance_network_function_map(
                    context, updated_network_function_map)
                self.nfp_notifier.update_network_function(
                    context.plugin_context, network_function_id,
                    context.current_node['config'])
            except Exception:
                LOG.exception("Update Network service Failed for "
                              "network function: %(nf_id)s",
                             {'nf_id': network_function_id})
        else:
            LOG.info("No action to take on update")

    def _get_service_chain_specs(self, context):
        current_specs = context.relevant_specs
        for spec in current_specs:
            filters = {'id': spec['nodes']}
            nodes = context.sc_plugin.get_servicechain_nodes(
                context.plugin_context, filters)
            for node in nodes:
                profile = context.sc_plugin.get_service_profile(
                    context.plugin_context, node['service_profile_id'])
                node['sc_service_profile'] = profile
            spec['sc_nodes'] = nodes
        return current_specs

    def _sc_head_gateway_node_service_targets(self, context,
                                              service_type, relationships):
        current_specs = context.relevant_specs
        service_targets = []
        for spec in current_specs:
            filters = {'id': spec['nodes']}
            nodes = context.sc_plugin.get_servicechain_nodes(
                context.plugin_context, filters)
            for node in nodes:
                profile = context.sc_plugin.get_service_profile(
                    context.plugin_context, node['service_profile_id'])
                if (profile['service_type'] != service_type and
                        profile['service_type'] in GATEWAY_PLUMBER_TYPE):
                    for relationship in relationships:
                        service_targets.extend(ncp_model.get_service_targets(
                            context.session,
                            servicechain_instance_id=context.instance['id'],
                            servicechain_node_id=node['id'],
                            relationship=relationship))
        return service_targets

    def _get_service_targets(self, context):
        service_type = context.current_profile['service_type']
        provider_service_targets = []
        consumer_service_targets = []
        service_flavor_str = context.current_profile['service_flavor']
        service_details = self._parse_service_flavor_string(service_flavor_str)
        nfp_context = NFPContext.get_nfp_context(context.instance['id'])
        is_gateway_type = False
        global GATEWAY_PLUMBER_TYPE
        if service_type in GATEWAY_PLUMBER_TYPE:
            for gateway_node in nfp_context['sc_gateway_type_nodes']:
                if gateway_node['context']:
                    service_target_info = gateway_node['context']
                    return service_target_info
            is_gateway_type = True

        service_targets = context.get_service_targets()
        # Bug with NCP. For create, its not setting service targets in context

        if not service_targets:
            service_targets = context.get_service_targets(update=True)

        if not service_targets and is_gateway_type:
            relationships = [nfp_constants.PROVIDER, nfp_constants.CONSUMER]
            service_targets = self._sc_head_gateway_node_service_targets(
                context,
                service_type,
                relationships)

        for service_target in service_targets:
            if service_target.relationship == nfp_constants.CONSUMER:
                consumer_service_targets.append(service_target)
            elif service_target.relationship == nfp_constants.PROVIDER:
                provider_service_targets.append(service_target)

        LOG.debug("provider targets: %s consumer targets %s", (
            provider_service_targets, consumer_service_targets))
        if (service_details['device_type'] != 'None' and (
            not provider_service_targets or (service_type in
            [pconst.FIREWALL, pconst.VPN] and not consumer_service_targets))):
            LOG.error("Service Targets are not created for the Node "
                      "of service_type %(service_type)s",
                      {'service_type': service_type})
            raise Exception(_("Service Targets are not created "
                              "for the Node"))

        if (not consumer_service_targets and
                not provider_service_targets):
            return {}

        service_target_info = {
            'provider_ports': [],
            'provider_subnet': None,
            'provider_pts': [],
            'provider_pt_objs': [],
            'provider_ptg': [],
            'consumer_ports': [],
            'consumer_subnet': None,
            'consumer_pts': [],
            'consumer_pt_objs': [],
            'consumer_ptg': []}

        for service_target in provider_service_targets:
            policy_target = context.gbp_plugin.get_policy_target(
                context.plugin_context, service_target.policy_target_id)
            policy_target_group = context.gbp_plugin.get_policy_target_group(
                context.plugin_context,
                policy_target['policy_target_group_id'])
            port = context.core_plugin.get_port(
                context.plugin_context, policy_target['port_id'])
            port['ip_address'] = port['fixed_ips'][0]['ip_address']
            subnet = context.core_plugin.get_subnet(
                context.plugin_context, port['fixed_ips'][0]['subnet_id'])
            service_target_info['provider_ports'].append(port)
            service_target_info['provider_subnet'] = subnet
            service_target_info['provider_pts'].append(policy_target['id'])
            service_target_info['provider_pt_objs'].append(policy_target)
            service_target_info['provider_ptg'].append(policy_target_group)

        for service_target in consumer_service_targets:
            policy_target = context.gbp_plugin.get_policy_target(
                context.plugin_context, service_target.policy_target_id)
            policy_target_group = context.gbp_plugin.get_policy_target_group(
                context.plugin_context,
                policy_target['policy_target_group_id'])
            port = context.core_plugin.get_port(
                context.plugin_context, policy_target['port_id'])
            port['ip_address'] = port['fixed_ips'][0]['ip_address']
            subnet = context.core_plugin.get_subnet(
                context.plugin_context, port['fixed_ips'][0]['subnet_id'])
            service_target_info['consumer_ports'].append(port)
            service_target_info['consumer_subnet'] = subnet
            service_target_info['consumer_pts'].append(policy_target['id'])
            service_target_info['consumer_pt_objs'].append(policy_target)
            service_target_info['consumer_ptg'].append(policy_target_group)

        for gateway_node in nfp_context['sc_gateway_type_nodes']:
            if gateway_node['service_type'] == service_type:
                gateway_node['context'] = service_target_info
        NFPContext.store_nfp_context(context.instance['id'],
                                     **nfp_context)
        return service_target_info

    # Needs a better algorithm
    def _is_node_order_in_spec_supported(self, context):
        current_specs = context.relevant_specs
        service_type_list_in_chain = []
        node_list = []
        for spec in current_specs:
            node_list.extend(spec['nodes'])

        for node_id in node_list:
            node_info = context.sc_plugin.get_servicechain_node(
                context.plugin_context, node_id)
            profile = context.sc_plugin.get_service_profile(
                context.plugin_context, node_info['service_profile_id'])
            service_type_list_in_chain.append(profile['service_type'])

        if len(service_type_list_in_chain) != len(
            set(service_type_list_in_chain)):
            raise DuplicateServiceTypeInChain()

        allowed_chain_combinations = [
            [pconst.VPN],
            [pconst.VPN, pconst.FIREWALL],
            [pconst.VPN, pconst.FIREWALL, pconst.LOADBALANCERV2],
            [pconst.FIREWALL],
            [pconst.FIREWALL, pconst.LOADBALANCERV2],
            [pconst.LOADBALANCERV2]]

        if service_type_list_in_chain not in allowed_chain_combinations:
            raise InvalidNodeOrderInChain(
                    node_order=allowed_chain_combinations)

        NFPContext.store_nfp_context(context.instance['id'],
                                     sc_node_count=len(node_list))

    def _get_consumers_for_provider(self, context, provider):
        '''
        {
            consuming_ptgs_details: [{'ptg': <>, 'subnets': <>}]
            consuming_eps_details: []
        }
        '''

        consuming_ptgs = []
        consuming_ptgs_details = []
        consuming_eps_details = []

        filters = {'id': provider['provided_policy_rule_sets']}
        provided_prs = context.gbp_plugin.get_policy_rule_sets(
            context.plugin_context, filters=filters)
        redirect_prs = None
        for prs in provided_prs:
            filters = {'id': prs['policy_rules']}
            policy_rules = context.gbp_plugin.get_policy_rules(
                context.plugin_context, filters=filters)
            for policy_rule in policy_rules:
                filters = {'id': policy_rule['policy_actions'],
                           'action_type': [gconst.GP_ACTION_REDIRECT]}
                policy_actions = context.gbp_plugin.get_policy_actions(
                    context.plugin_context, filters=filters)
                if policy_actions:
                    redirect_prs = prs
                    break

        if not redirect_prs:
            LOG.error("Redirect rule doesn't exist in policy target rule "
                      " set")
            return consuming_ptgs_details, consuming_eps_details

        consuming_ptg_ids = redirect_prs['consuming_policy_target_groups']
        consuming_ep_ids = redirect_prs['consuming_external_policies']
        if consuming_ptg_ids:
            consuming_ptgs = context.gbp_plugin.get_policy_target_groups(
                context.plugin_context, filters={'id': consuming_ptg_ids})
        if consuming_ep_ids:
            consuming_eps_details = context.gbp_plugin.get_external_policies(
                context.plugin_context, filters={'id': consuming_ep_ids})

        for ptg in consuming_ptgs:
            subnet_ids = ptg['subnets']
            subnets = context.core_plugin.get_subnets(
                context.plugin_context, filters={'id': subnet_ids})
            consuming_ptgs_details.append({'ptg': ptg, 'subnets': subnets})

        return consuming_ptgs_details, consuming_eps_details

    def _get_nfp_network_function(self, context):
        """
        nfp_nf_data :-

        {'resource_owner_context': <>,
         'service_chain_instance': <>,
         'service_chain_node': <>,
         'service_profile': <>,
         'service_config': context.current_node.get('config'),
         'provider': {'pt':<>, 'ptg':<>, 'port':<>, 'subnet':<>},
         'consumer': {'pt':<>, 'ptg':<>, 'port':<>, 'subnet':<>},
         'management': {'pt':<>, 'ptg':<>, 'port':<>, 'subnet':<>},
         'management_ptg_id': <>,
         'network_function_mode': nfp_constants.GBP_MODE,
         'tenant_id': <>,
         'consuming_ptgs_details': [],
         'consuming_eps_details': []
        }

        """
        nfp_nf_data = {}

        sc_instance = context.instance
        service_targets = self._get_service_targets(context)

        consuming_ptgs_details = []
        consuming_eps_details = []
        if service_targets:
            consuming_ptgs_details, consuming_eps_details = \
                self._get_consumers_for_provider(context,
                    service_targets['provider_ptg'][0])

        if context.current_profile['service_type'] == pconst.LOADBALANCERV2:
            config_param_values = sc_instance.get('config_param_values', {})
            if config_param_values:
                config_param_values = jsonutils.loads(config_param_values)
            vip_ip = config_param_values.get('vip_ip')
            if not vip_ip:
                raise VipNspNotSetonProvider()

            if service_targets:
                for provider_port in service_targets['provider_ports']:
                    provider_port['allowed_address_pairs'] = [
                        {'ip_address': vip_ip}]
                    port = {
                        'port': provider_port
                    }
                    context.core_plugin.update_port(
                        context.plugin_context, provider_port['id'], port)

        provider = {
            'pt': service_targets.get('provider_pt_objs', []),
            'ptg': service_targets.get('provider_ptg', []),
            'port': service_targets.get('provider_ports', []),
            'subnet': service_targets.get('provider_subnet', None),
            'port_model': nfp_constants.GBP_PORT,
            'port_classification': nfp_constants.PROVIDER}

        consumer_pt = None
        consumer_ptg = None
        consumer_ports = None

        if service_targets.get('consumer_pt_objs'):
            consumer_pt = service_targets.get('consumer_pt_objs')
        if service_targets.get('consumer_ptg'):
            consumer_ptg = service_targets.get('consumer_ptg')
        if service_targets.get('consumer_ports'):
            consumer_ports = service_targets.get('consumer_ports')

        consumer = {
            'pt': consumer_pt,
            'ptg': consumer_ptg,
            'port': consumer_ports,
            'subnet': service_targets.get('consumer_subnet', None),
            'port_model': nfp_constants.GBP_PORT,
            'port_classification': nfp_constants.CONSUMER}

        management = {
            'pt': None,
            'ptg': None,
            'port': None,
            'subnet': None,
            'port_model': nfp_constants.GBP_NETWORK,
            'port_classification': nfp_constants.MANAGEMENT}

        service_chain_specs = self._get_service_chain_specs(context)

        nfp_nf_data = {
            'resource_owner_context': context._plugin_context.to_dict(),
            'service_chain_instance': sc_instance,
            'service_chain_node': context.current_node,
            'service_profile': context.current_profile,
            'service_config': context.current_node.get('config'),
            'provider': provider,
            'consumer': consumer,
            'management': management,
            'management_ptg_id': sc_instance['management_ptg_id'],
            'network_function_mode': nfp_constants.GBP_MODE,
            'tenant_id': context.provider['tenant_id'],
            'consuming_ptgs_details': consuming_ptgs_details,
            'consuming_eps_details': consuming_eps_details,
            'service_chain_specs': service_chain_specs}
        return nfp_nf_data

    def _create_network_function(self, context):
        nfp_create_nf_data = self._get_nfp_network_function(context)
        LOG.info("Received Call CREATE NETWORK FUNCTION for tenant: "
                 "%(tenant_id)s with service profile:"
                 "%(service_profile)s",
                 {'tenant_id': nfp_create_nf_data['tenant_id'],
                  'service_profile': nfp_create_nf_data['service_profile']})
        self.nfp_notifier.create_network_function(
            context.plugin_context, nfp_create_nf_data)

    def _delete_network_function(self, context, network_function_id):
        nfp_delete_nf_data = self._get_nfp_network_function(context)
        self._detach_port_from_pts(context,
                                   nfp_delete_nf_data['provider']['pt'])
        if nfp_delete_nf_data['consumer'].get('pt'):
            self._detach_port_from_pts(context,
                                       nfp_delete_nf_data['consumer']['pt'])
        LOG.info("Received Call DELETE NETWORK FUNCTION for tenant: "
                 "%(tenant_id)s with service profile:"
                 "%(service_profile)s",
                 {'tenant_id': nfp_delete_nf_data['tenant_id'],
                  'service_profile': nfp_delete_nf_data['service_profile']})
        self.nfp_notifier.delete_network_function(
            context.plugin_context, network_function_id, nfp_delete_nf_data)

    def _detach_port_from_pts(self, context, policy_targets):
        '''
        Disassociate port from policy target
        '''
        try:
            for pt in policy_targets:
                context.gbp_plugin.update_policy_target(
                        context.plugin_context, pt['id'],
                        {'policy_target': {'port_id': None}})
                LOG.debug('Detached port %(port)s from pt %(pt)s',
                          {'pt': pt['id'], 'port': pt['port_id']})

        except Exception:
            LOG.warning("Failed to disassociate port from"
                " pt: %(pt)s, Error: %(exc)s", {'pt': pt, 'exc': exc})

    def _update_ptg(self, context):
        if hasattr(context, 'provider') and context.provider['description']:
            gateway_desc = 'opflex_eoc' in context.provider[
                'description'].split(':')
            if gateway_desc:
                pts = context.gbp_plugin.get_policy_targets(
                        context.plugin_context,
                        filters={'port_id': [context.provider[
                            'description'].split(':')][-1]})
                (pt,) = pts
                filters = {'description': [context.provider['description']]}
                ptgs = context.gbp_plugin.get_policy_target_groups(
                        context.plugin_context, filters)
                prs = []
                for ptg in ptgs:
                    prs += ptg['provided_policy_rule_sets']
                context.gbp_plugin.update_policy_target_group(
                        context.plugin_context,
                        pt['policy_target_group_id'],
                        {'policy_target_group':
                            {'provided_policy_rule_sets':
                                dict((x, '') for x in prs)}})



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock

# from aim.api import infra as aim_infra
from aim.api import resource as aim_res
from gbpservice.neutron.tests.unit.services.grouppolicy import (
    test_aim_mapping_driver as test_aim_base)

from neutron.objects.qos import policy as policy_object
from neutron.objects.qos import rule as rule_object
from neutron_lib import context
from neutron_lib.plugins import directory
from oslo_log import log as logging
from oslo_utils import uuidutils

from gbpservice.neutron.services.grouppolicy import config

LOG = logging.getLogger(__name__)


class TestAIMQosBase(test_aim_base.AIMBaseTestCase):

    def setUp(self, *args, **kwargs):
        net_type = kwargs.pop('tenant_network_types', None) or ['vlan']
        ml2_options = {'mechanism_drivers': ['apic_aim', 'openvswitch'],
                       'extension_drivers': ['apic_aim', 'port_security',
                                             'dns', 'qos'],
                       'type_drivers': ['opflex', 'local', 'vlan'],
                       'tenant_network_types': net_type}
        config.cfg.CONF.set_override(
            'network_vlan_ranges', ['physnet1:100:200'], group='ml2_type_vlan')
        self._aim_mech_driver = None
        self._qos_driver = None
        super(TestAIMQosBase, self).setUp(
            *args, ml2_options=ml2_options, qos_plugin='qos', **kwargs)
        self._plugin = directory.get_plugin()
        self._plugin.remove_networks_from_down_agents = mock.Mock()
        self._plugin.is_agent_down = mock.Mock(return_value=False)
        self._ctx = context.get_admin_context()
        self.ctxt = context.Context('fake_user', 'fake_tenant')
        self.admin_ctxt = context.get_admin_context()

        self.policy_data = {
            'policy': {'project_id': uuidutils.generate_uuid(),
                       'name': 'test-policy',
                       'description': 'Test policy description',
                       'shared': True,
                       'is_default': False}}

        self.rule_data = {
            'egress_bandwidth_limit_rule': {
                'id': uuidutils.generate_uuid(),
                'direction': 'egress',
                'max_kbps': 100,
                'max_burst_kbps': 150},
            'ingress_bandwidth_limit_rule': {
                'id': uuidutils.generate_uuid(),
                'direction': 'ingress',
                'max_kbps': 101,
                'max_burst_kbps': 1150},
            'dscp_marking_rule': {'id': uuidutils.generate_uuid(),
                                  'dscp_mark': 16}, }

        self.egress_rule = rule_object.QosBandwidthLimitRule(
            self.ctxt, **self.rule_data['egress_bandwidth_limit_rule'])

        self.ingress_rule = rule_object.QosBandwidthLimitRule(
            self.ctxt, **self.rule_data['ingress_bandwidth_limit_rule'])

        self.dscp_rule = rule_object.QosDscpMarkingRule(
            self.ctxt, **self.rule_data['dscp_marking_rule'])

    def tearDown(self):
        super(TestAIMQosBase, self).tearDown()

    @property
    def qos_plugin(self):
        if not self._qos_plugin:
            self._qos_plugin = directory.get_plugin('qos')
        return self._qos_plugin

    @property
    def qos_driver(self):
        if not self._qos_driver:
            self._qos_driver = self.aim_mech.qos_driver
        return self._qos_driver

    @property
    def aim_mech(self):
        if not self._aim_mech_driver:
            self._aim_mech_driver = (
                self._plugin.mechanism_manager.mech_drivers['apic_aim'].obj)
        return self._aim_mech_driver


class TestQosPolicy(TestAIMQosBase):

    def test_create_delete_update_policy(self):
        _policy = policy_object.QosPolicy(
            self.ctxt, **self.policy_data['policy'])
        _policy.create()
        self.qos_driver.create_policy_precommit(self.ctxt, _policy)
        tenant_name = 'prj_' + self.ctxt.tenant_id
        pol = aim_res.QosRequirement(name=_policy.id, tenant_name=tenant_name)
        pol = self.aim_mgr.get(self._aim_context, pol)
        self.assertIsNotNone(pol)

        # Test dscp rule
        setattr(_policy, "rules", [self.dscp_rule])
        self.qos_driver.update_policy_precommit(self.ctxt, _policy)
        pol = self.aim_mgr.get(self._aim_context, pol)
        self.assertIsNotNone(pol)
        self.assertEqual(pol.dscp,
                         self.rule_data['dscp_marking_rule']['dscp_mark'])

        # Test bw rules
        setattr(_policy, "rules", [self.egress_rule, self.ingress_rule])
        self.qos_driver.update_policy_precommit(self.ctxt, _policy)
        pol = self.aim_mgr.get(self._aim_context, pol)
        self.assertIsNotNone(pol)
        self.assertEqual(pol.ingress_dpp_pol, self.ingress_rule.id)
        self.assertEqual(pol.egress_dpp_pol, self.egress_rule.id)
        self.assertIsNone(pol.dscp)
        egress_bw_rule = aim_res.QosDppPol(
            name=self.egress_rule.id, tenant_name=tenant_name)
        egress_bw_rule = self.aim_mgr.get(self._aim_context, egress_bw_rule)
        self.assertIsNotNone(egress_bw_rule)
        self.assertEqual(egress_bw_rule.burst,
                         str(self.egress_rule.max_burst_kbps))
        self.assertEqual(egress_bw_rule.rate, self.egress_rule.max_kbps)
        ingress_bw_rule = aim_res.QosDppPol(
            name=self.ingress_rule.id, tenant_name=tenant_name)
        ingress_bw_rule = self.aim_mgr.get(self._aim_context, ingress_bw_rule)
        self.assertIsNotNone(ingress_bw_rule)
        self.assertEqual(ingress_bw_rule.burst,
                         str(self.ingress_rule.max_burst_kbps))
        self.assertEqual(ingress_bw_rule.rate, self.ingress_rule.max_kbps)

        # Clean up the rules
        setattr(_policy, "rules", [])
        self.qos_driver.update_policy_precommit(self.ctxt, _policy)
        pol = self.aim_mgr.get(self._aim_context, pol)
        self.assertIsNone(pol.egress_dpp_pol)
        self.assertIsNone(pol.ingress_dpp_pol)
        egress_bw_rule = self.aim_mgr.get(self._aim_context, egress_bw_rule)
        self.assertIsNone(egress_bw_rule)
        ingress_bw_rule = self.aim_mgr.get(self._aim_context, ingress_bw_rule)
        self.assertIsNone(ingress_bw_rule)

        self.qos_driver.delete_policy_precommit(self.ctxt, _policy)
        pol = self.aim_mgr.get(self._aim_context, pol)
        self.assertIsNone(pol)

    def _make_qos_policy(self):
        qos_policy = policy_object.QosPolicy(
            self.admin_ctxt, **self.policy_data['policy'])
        qos_policy.create()
        return qos_policy

    def test_attach_detach_network_qos(self):
        net_qos_obj = self._make_qos_policy()
        net_qos_id = net_qos_obj.id if net_qos_obj else None
        network = self._make_network(self.fmt, 'net1', True,
                                     arg_list=tuple(list(['qos_policy_id'])),
                                     qos_policy_id=net_qos_id)
        epg = self.aim_mech._get_epg_by_network_id(self._ctx.session,
                                                   network['network']['id'])
        epg = self.aim_mgr.get(self._aim_context, epg)
        self.assertEqual(epg.qos_name, net_qos_id)

        data = {'network': {'qos_policy_id': None}}
        self._update('networks', network['network']['id'], data)
        epg = self.aim_mgr.get(self._aim_context, epg)
        self.assertIsNone(epg.qos_name)

    def _test_invalid_network_exception(self, kwargs):
        # Verify creating network with QoS fails
        net_qos_obj = self._make_qos_policy()
        net_qos_id = net_qos_obj.id if net_qos_obj else None
        kwargs['qos_policy_id'] = net_qos_id

        resp = self._create_network(
            self.fmt, 'net', True, arg_list=tuple(list(kwargs.keys())),
            **kwargs)
        result = self.deserialize(self.fmt, resp)
        self.assertEqual(
            'InvalidNetworkForQos',
            result['NeutronError']['type'])

    def test_external_network_exception(self):
        self._test_invalid_network_exception({'router:external': True})

    def test_svi_network_exception(self):
        self._test_invalid_network_exception({'apic:svi': True})



# Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Service chain ordering

Revision ID: e8005b9b1efc

"""

# revision identifiers, used by Alembic.
revision = 'e8005b9b1efc'
down_revision = '8e14fcb1587e'


from alembic import op
import sqlalchemy as sa


def upgrade():

    op.add_column(
        'sc_spec_node_associations',
        sa.Column('position', sa.Integer(), nullable=False)
    )
    op.add_column(
        'sc_instance_spec_mappings',
        sa.Column('position', sa.Integer(), nullable=False)
    )


def downgrade():

    op.drop_column('sc_instance_spec_mappings', 'position')
    op.drop_column('sc_spec_node_associations', 'position')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ast
import copy

from gbpservice.contrib.nfp.config_orchestrator.common import common
from gbpservice.nfp.common import constants as const
from gbpservice.nfp.common import data_formatter as df
from gbpservice.nfp.common import utils
from gbpservice.nfp.core import context as module_context
from gbpservice.nfp.core import log as nfp_logging
from gbpservice.nfp.lib import transport

from neutron_vpnaas.db.vpn import vpn_db

from oslo_log import helpers as log_helpers
import oslo_messaging as messaging

LOG = nfp_logging.getLogger(__name__)

"""
RPC handler for VPN service
"""


class VpnAgent(vpn_db.VPNPluginDb, vpn_db.VPNPluginRpcDbMixin):
    RPC_API_VERSION = '1.0'
    target = messaging.Target(version=RPC_API_VERSION)

    def __init__(self, conf, sc):
        super(VpnAgent, self).__init__()
        self._conf = conf
        self._sc = sc
        self._db_inst = super(VpnAgent, self)

    def _get_vpn_context(self, context, tenant_id, vpnservice_id,
                         ikepolicy_id, ipsecpolicy_id,
                         ipsec_site_conn_id, desc):
        vpnservices = self._get_vpnservices(context, tenant_id,
                                            vpnservice_id, desc)
        ikepolicies = self._get_ikepolicies(context, tenant_id,
                                            ikepolicy_id)
        ipsecpolicies = self._get_ipsecpolicies(context, tenant_id,
                                                ipsecpolicy_id)
        ipsec_site_conns = self._get_ipsec_site_conns(context, tenant_id,
                                                      ipsec_site_conn_id, desc)

        return {'vpnservices': vpnservices,
                'ikepolicies': ikepolicies,
                'ipsecpolicies': ipsecpolicies,
                'ipsec_site_conns': ipsec_site_conns}

    def _context(self, context, tenant_id, resource, resource_data):
        if context.is_admin:
            tenant_id = context.tenant_id
        if resource.lower() == 'ipsec_site_connection':
            vpn_ctx_db = self._get_vpn_context(context,
                                               tenant_id,
                                               resource_data[
                                                   'vpnservice_id'],
                                               resource_data[
                                                   'ikepolicy_id'],
                                               resource_data[
                                                   'ipsecpolicy_id'],
                                               resource_data['id'],
                                               resource_data[
                                                   'description'])
            return vpn_ctx_db
        elif resource.lower() == 'vpn_service':
            return {'vpnservices': [resource_data]}
        else:
            return None

    def _prepare_resource_context_dicts(self, context, tenant_id,
                                        resource, resource_data,
                                        context_resource_data):
        # Prepare context_dict
        ctx_dict = context.to_dict()
        # Collecting db entry required by configurator.
        # Addind service_info to neutron context and sending
        # dictionary format to the configurator.
        db = self._context(context, tenant_id, resource,
                           resource_data)
        rsrc_ctx_dict = copy.deepcopy(ctx_dict)
        rsrc_ctx_dict.update({'service_info': db})
        rsrc_ctx_dict.update({'resource_data': context_resource_data})
        return ctx_dict, rsrc_ctx_dict

    def _get_resource_data(self, description, resource_type):
        resource_data = df.get_network_function_info(description,
                                                     resource_type)
        return resource_data

    def _update_request_data(self, body, description):
        pass

    def _data_wrapper(self, context, tenant_id, nf, **kwargs):
        nfp_context = {}
        description, str_description = (
            utils.get_vpn_description_from_nf(nf))
        description.update({'tenant_id': tenant_id})
        context_resource_data = self._get_resource_data(description,
                                                        const.VPN)
        resource = kwargs['rsrc_type']
        resource_data = kwargs['resource']
        # REVISIT(dpak): We need to avoid resource description
        # dependency in OTC and instead use neutron context description.
        resource_data['description'] = str_description
        if resource.lower() == 'ipsec_site_connection':
            nfp_context = {'network_function_id': nf['id'],
                           'ipsec_site_connection_id': kwargs[
                               'rsrc_id']}

        ctx_dict, rsrc_ctx_dict = self.\
            _prepare_resource_context_dicts(context, tenant_id,
                                            resource, resource_data,
                                            context_resource_data)
        service_vm_context = utils.get_service_vm_context(
                                                description['service_vendor'])
        nfp_context.update({'neutron_context': ctx_dict,
                            'service_vm_context': service_vm_context,
                            'requester': 'nas_service',
                            'logging_context':
                                module_context.get()['log_context']})
        resource_type = 'vpn'
        kwargs.update({'neutron_context': rsrc_ctx_dict})
        body = common.prepare_request_data(nfp_context, resource,
                                           resource_type, kwargs,
                                           description['service_vendor'])
        self._update_request_data(body, description)
        return body

    def _fetch_nf_from_resource_desc(self, desc):
        desc_dict = ast.literal_eval(desc)
        nf_id = desc_dict['network_function_id']
        return nf_id

    @log_helpers.log_method_call
    def vpnservice_updated(self, context, **kwargs):
        nfp_context = module_context.init()
        LOG.info("Received RPC VPN SERVICE UPDATED with data:%(data)s",
                 {'data': kwargs})
        # Fetch nf_id from description of the resource
        nf_id = self._fetch_nf_from_resource_desc(kwargs[
            'resource']['description'])
        nfp_context['log_context']['meta_id'] = nf_id
        nf = common.get_network_function_details(context, nf_id)
        reason = kwargs['reason']
        body = self._data_wrapper(context, kwargs[
            'resource']['tenant_id'], nf, **kwargs)
        transport.send_request_to_configurator(self._conf,
                                               context, body,
                                               reason)

    def _proxy_subnet_cidr(self, description):
        tokens = description.split(';')
        return tokens[5].split('=')[1]

    def _get_vpnservices(self, context, tenant_id, vpnservice_id, desc):
        filters = {'tenant_id': [tenant_id],
                   'id': [vpnservice_id]}
        args = {'context': context, 'filters': filters}
        vpnservices = self._db_inst.get_vpnservices(**args)
        for vpnservice in vpnservices:
            vpnservice['description'] = desc
        return vpnservices

    def _get_ikepolicies(self, context, tenant_id, ikepolicy_id):
        filters = {'tenant_id': [tenant_id],
                   'id': [ikepolicy_id]}
        args = {'context': context, 'filters': filters}
        return self._db_inst.get_ikepolicies(**args)

    def _get_ipsecpolicies(self, context, tenant_id, ipsecpolicy_id):
        filters = {'tenant_id': [tenant_id],
                   'id': [ipsecpolicy_id]}
        args = {'context': context, 'filters': filters}
        return self._db_inst.get_ipsecpolicies(**args)

    def _get_ipsec_site_conns(self, context, tenant_id, ipsec_site_conn_id,
                              desc):
        filters = {'tenant_id': [tenant_id],
                   'id': [ipsec_site_conn_id]}
        args = {'context': context, 'filters': filters}
        ipsec_site_conns = self._db_inst.get_ipsec_site_connections(**args)
        for ipsec_site_conn in ipsec_site_conns:
            ipsec_site_conn['description'] = desc
        return ipsec_site_conns



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from neutron.tests.unit import testlib_api
from neutron_lib import context
from oslo_db import exception as exc
from oslo_utils import importutils

from gbpservice.neutron.db import all_models  # noqa
from gbpservice.neutron.plugins.ml2plus.drivers.apic_aim import db

DB_PLUGIN_KLASS = 'neutron.db.db_base_plugin_v2.NeutronDbPluginV2'


class PortToHAIPAddressBindingTestCase(testlib_api.SqlTestCase):

    def setUp(self):
        super(PortToHAIPAddressBindingTestCase, self).setUp()
        self.plugin = importutils.import_object(DB_PLUGIN_KLASS)
        self.context = context.get_admin_context()
        self.net1_data = {'network': {'id': 'fake-net1-id',
                                      'name': 'net1',
                                      'admin_state_up': True,
                                      'tenant_id': 'test-tenant',
                                      'shared': False}}
        self.net2_data = {'network': {'id': 'fake-net2-id',
                                      'name': 'net2',
                                      'admin_state_up': True,
                                      'tenant_id': 'test-tenant',
                                      'shared': False}}
        self.port1_data = {'port': {'id': 'fake-port1-id',
                                    'name': 'port1',
                                    'network_id': 'fake-net1-id',
                                    'tenant_id': 'test-tenant',
                                    'device_id': 'fake_device',
                                    'device_owner': 'fake_owner',
                                    'fixed_ips': [],
                                    'mac_address': 'fake-mac',
                                    'admin_state_up': True}}
        # Port that is in the same network as port_1
        self.port1_2_data = {'port': {'id': 'fake-port1-2-id',
                                      'name': 'port1',
                                      'network_id': 'fake-net1-id',
                                      'tenant_id': 'test-tenant',
                                      'device_id': 'fake_device',
                                      'device_owner': 'fake_owner',
                                      'fixed_ips': [],
                                      'mac_address': 'fake-mac-2',
                                      'admin_state_up': True}}
        self.port2_data = {'port': {'id': 'fake-port2-id',
                                    'name': 'port2',
                                    'network_id': 'fake-net2-id',
                                    'tenant_id': 'test-tenant',
                                    'device_id': 'fake_device',
                                    'device_owner': 'fake_owner',
                                    'fixed_ips': [],
                                    'mac_address': 'fake-mac',
                                    'admin_state_up': True}}
        self.ha_ip1 = "ha-ip-1"
        self.ha_ip2 = "ha-ip-2"
        self.plugin.create_network(self.context, self.net1_data)
        self.plugin.create_network(self.context, self.net2_data)
        self.port1 = self.plugin.create_port(self.context, self.port1_data)
        self.port1_2 = self.plugin.create_port(self.context, self.port1_2_data)
        self.port2 = self.plugin.create_port(self.context, self.port2_data)
        self.port_haip = db.DbMixin()
        self.port_haip.plugin = self.plugin

    def test_set_and_get_port_to_ha_ip_binding(self):
        # Test new HA IP address to port binding can be created
        obj = self.port_haip.set_port_id_for_ha_ipaddress(
            self.port1['id'], self.ha_ip1)
        self.assertEqual(self.port1['id'], obj['port_id'])
        self.assertEqual(self.ha_ip1, obj['ha_ip_address'])
        # In this test case we also test that same HA IP address can be set/get
        # for two different ports in different networks
        obj = self.port_haip.set_port_id_for_ha_ipaddress(
            self.port2['id'], self.ha_ip1)
        self.assertEqual(self.port2['id'], obj['port_id'])
        self.assertEqual(self.ha_ip1, obj['ha_ip_address'])
        # Test get
        obj = self.port_haip.get_port_for_ha_ipaddress(
            self.ha_ip1, self.port1['network_id'])
        self.assertEqual(self.port1['id'], obj['port_id'])
        obj = self.port_haip.get_port_for_ha_ipaddress(
            self.ha_ip1, self.port2['network_id'])
        self.assertEqual(self.port2['id'], obj['port_id'])

    def test_port_to_multiple_ha_ip_binding(self):
        self.port_haip.set_port_id_for_ha_ipaddress(self.port1['id'],
                                                    self.ha_ip1)
        self.port_haip.set_port_id_for_ha_ipaddress(self.port1['id'],
                                                    self.ha_ip2)
        obj = self.port_haip.get_port_for_ha_ipaddress(
            self.ha_ip1, self.port1['network_id'])
        self.assertEqual(self.port1['id'], obj['port_id'])
        obj = self.port_haip.get_port_for_ha_ipaddress(
            self.ha_ip2, self.port1['network_id'])
        self.assertEqual(self.port1['id'], obj['port_id'])

    def test_delete_port_for_ha_ip_binding(self):
        self.port_haip.set_port_id_for_ha_ipaddress(self.port1['id'],
                                                    self.ha_ip1)
        result = self.port_haip.delete_port_id_for_ha_ipaddress(
            self.port1['id'], self.ha_ip1)
        self.assertEqual(1, result)
        obj = self.port_haip.get_port_for_ha_ipaddress(
            self.ha_ip1, self.port2['network_id'])
        self.assertIsNone(obj)

    def test_get_ha_ip_addresses_for_port(self):
        self.port_haip.set_port_id_for_ha_ipaddress(self.port1['id'],
                                                    self.ha_ip1)
        self.port_haip.set_port_id_for_ha_ipaddress(self.port1['id'],
                                                    self.ha_ip2)
        ha_ips = self.port_haip.get_ha_ipaddresses_for_port(self.port1['id'])
        self.assertEqual(sorted([self.ha_ip1, self.ha_ip2]), ha_ips)

    def test_idempotent(self):
        self.port_haip.set_port_id_for_ha_ipaddress(self.port1['id'],
                                                    self.ha_ip1)
        obj = self.port_haip.set_port_id_for_ha_ipaddress(self.port1['id'],
                                                          self.ha_ip1)
        self.assertEqual(self.port1['id'], obj['port_id'])
        self.assertEqual(self.ha_ip1, obj['ha_ip_address'])

    def test_set_non_existing_port(self):
        self.assertRaises(exc.DBReferenceError,
                          self.port_haip.set_port_id_for_ha_ipaddress,
                          "fake", self.ha_ip1)

    def test_delete_non_existing_entry(self):
        self.port_haip.set_port_id_for_ha_ipaddress(self.port1['id'],
                                                    self.ha_ip1)
        result = self.port_haip.delete_port_id_for_ha_ipaddress(
            self.port1['id'], "fake")
        self.assertEqual(0, result)
        result = self.port_haip.delete_port_id_for_ha_ipaddress("fake",
                                                                self.ha_ip1)
        self.assertEqual(0, result)

    def test_ip_owner_update(self):
        ip_owner_info = {'port': self.port1['id'],
                         'ip_address_v4': self.ha_ip1}

        # set new owner
        ports = self.port_haip.update_ip_owner(ip_owner_info)
        obj = self.port_haip.get_port_for_ha_ipaddress(
            self.ha_ip1, self.port1['network_id'])
        self.assertEqual(self.port1['id'], obj['port_id'])
        self.assertTrue(self.port1['id'] in ports)

        # update owner
        self.port2_data['port']['id'] = 'fake-port3-id'
        self.port2_data['port']['network_id'] = self.port1['network_id']
        self.port2_data['port']['mac_address'] = 'fake-mac-3'
        port3 = self.plugin.create_port(self.context, self.port2_data)

        ip_owner_info['port'] = port3['id']
        ports = self.port_haip.update_ip_owner(ip_owner_info)
        obj = self.port_haip.get_port_for_ha_ipaddress(
            self.ha_ip1, port3['network_id'])
        self.assertEqual(port3['id'], obj['port_id'])

    def test_ip_replaced(self):
        ip_owner_info = {'port': self.port1['id'],
                         'ip_address_v4': self.ha_ip1}
        self.port_haip.update_ip_owner(ip_owner_info)
        # Verify only one entry is there
        dump = self.port_haip.get_ha_port_associations()
        self.assertEqual(1, len(dump))
        self.assertEqual(self.port1['id'], dump[0].port_id)
        self.assertEqual(self.ha_ip1, dump[0].ha_ip_address)

        # Now override with port1_2
        ip_owner_info['port'] = self.port1_2['id']
        self.port_haip.update_ip_owner(ip_owner_info)
        # Verify still one entry exists
        dump = self.port_haip.get_ha_port_associations()
        self.assertEqual(1, len(dump))
        self.assertEqual(self.port1_2['id'], dump[0].port_id)
        self.assertEqual(self.ha_ip1, dump[0].ha_ip_address)

        # Override again, but with a different net_id to keep both records
        ip_owner_info['port'] = self.port1['id']
        ip_owner_info['network_id'] = 'new_net_id'
        self.port_haip.update_ip_owner(ip_owner_info)
        # Verify still one entry exists
        dump = self.port_haip.get_ha_port_associations()
        self.assertEqual(2, len(dump))

    def test_duplicate_entry_handled_gracefully(self):
        self.port_haip.set_port_id_for_ha_ipaddress(
            self.port1['id'], self.ha_ip1)
        # Set this twice, without hijacking the query
        obj = self.port_haip.set_port_id_for_ha_ipaddress(
            self.port1['id'], self.ha_ip1)
        self.assertEqual(obj.port_id, self.port1['id'])
        self.assertEqual(obj.ha_ip_address, self.ha_ip1)
        # Now simulate null return from query
        self.port_haip._get_ha_ipaddress = mock.Mock(return_value=None)
        obj = self.port_haip.set_port_id_for_ha_ipaddress(
            self.port1['id'], self.ha_ip1)
        self.assertIsNone(obj)



# Copyright (c) 2017 Cisco Systems
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from neutron_lib import context

from gbpservice.neutron.services.apic_aim import l3_plugin
from gbpservice.neutron.tests.unit.services.grouppolicy import (
    test_aim_mapping_driver)


TENANT = 'tenant1'
ROUTER = 'router1'
NETWORK = 'network1'
CIDR = '10.0.0.0/16'


class TestCiscoApicAimL3Plugin(test_aim_mapping_driver.AIMBaseTestCase):
    '''Test class for the Cisco APIC AIM L3 Plugin

       This is a set of tests specific to the Cisco APIC AIM
       L3 plugin. It currently derives from the AIMBaseTestCase
       class so that it can inherit test infrastructure from those classes.
    '''

    def _verify_event_handler_calls(self, data, expected_call_count=1):
        if not hasattr(data, '__iter__') or isinstance(data, dict):
            data = [data]
        self.assertEqual(expected_call_count, self.handler_mock.call_count)
        call_idx = -1
        for item in data:
            if item:
                model = self.handler_mock.call_args_list[call_idx][0][-1]
                self.assertEqual(model['id'], item['id'])
                self.assertEqual(model['tenant_id'], item['tenant_id'])
            call_idx = call_idx - 1

    def setUp(self):
        handler_patch = mock.patch(
            'neutron.quota.resource.TrackedResource._db_event_handler')
        self.handler_mock = handler_patch.start()
        def_sec_group_patch = mock.patch(
            'neutron.db.securitygroups_db.SecurityGroupDbMixin.'
            '_ensure_default_security_group')
        def_sec_group_patch.start()
        super(TestCiscoApicAimL3Plugin, self).setUp()
        get_sec_group_port_patch = mock.patch(
            'neutron.db.securitygroups_db.SecurityGroupDbMixin.'
            '_get_security_groups_on_port', return_value=set())
        get_sec_group_port_patch.start()

        # Set up L2 objects for L3 test
        attr = {'tenant_id': TENANT}
        resp = self._create_network(self.fmt, NETWORK, True, **attr)
        self.network = self.deserialize(self.fmt, resp)['network']
        attr.update({'network_id': self.network['id']})
        resp = self._create_subnet(self.fmt, self.network['id'], CIDR, **attr)
        self.subnet = self.deserialize(self.fmt, resp)['subnet']
        resp = self._create_port(self.fmt, self.network['id'],
                                 tenant_id=TENANT)
        self.port = self.deserialize(self.fmt, resp)['port']
        self.interface_info = {'subnet': {'subnet_id': self.subnet['id']},
                               'port': {'port_id': self.port['id']}}
        self.context = context.get_admin_context()
        self.context.tenant_id = TENANT

        self.plugin = l3_plugin.ApicL3Plugin()

    def _test_add_router_interface(self, interface_info):

        attr = {'router': {'tenant_id': TENANT,
                           'admin_state_up': True,
                           'name': ROUTER}}
        router = self.plugin.create_router(self.context, attr)

        with mock.patch('neutron_lib.callbacks.registry.notify'):
            info = self.plugin.add_router_interface(self.context,
                                                    router['id'],
                                                    interface_info)
            self.assertEqual(info['id'], router['id'])
            self.assertEqual(info['tenant_id'], TENANT)
            if interface_info.get('port_id'):
                self.assertEqual(info['port_id'], self.port['id'])
            else:
                self.assertNotEqual(info['port_id'], self.port['id'])
            self.assertEqual(info['subnet_id'], self.subnet['id'])
            self.assertEqual(info['network_id'], self.network['id'])

    def _test_remove_router_interface(self, interface_info):
        with mock.patch('neutron.db.l3_db.'
                        'L3_NAT_db_mixin.remove_router_interface') as if_mock:
            self.plugin.remove_router_interface(self.context, ROUTER,
                                                interface_info)
            self.assertEqual(1, if_mock.call_count)

    def test_add_router_interface_subnet(self):
        self._test_add_router_interface(self.interface_info['subnet'])

    def test_add_router_interface_port(self):
        self._test_add_router_interface(self.interface_info['port'])

    def test_remove_router_interface_subnet(self):
        self._test_remove_router_interface(self.interface_info['subnet'])

    def test_remove_router_interface_port(self):
        self._test_remove_router_interface(self.interface_info['port'])

    def test_create_delete_floating_ip_triggers_event(self):
        net = self._make_network('json', 'meh', True)
        subnet = self._make_subnet('json', net, '14.0.0.1',
                                   '14.0.0.0/24')['subnet']
        self._set_net_external(subnet['network_id'])
        kwargs = {'floatingip': {'tenant_id': TENANT,
                                 'subnet_id': subnet['id'],
                                 'floating_network_id': subnet['network_id']}}
        self.handler_mock.reset_mock()
        floatingip = self.plugin.create_floatingip(self.context, kwargs)
        internal_ports = self._show('ports',
                                    floatingip['port_id'])['ports']
        for port in internal_ports:
            if port['device_owner'] == 'network:floatingip' and (
                    port['device_id'] == floatingip['id']):
                internal_port = port
        # When a floatingip is created it also creates port, therefore
        # there will be four calls in total to the event handler
        self._verify_event_handler_calls(floatingip,
                                         expected_call_count=2)
        self._delete('floatingips', floatingip['id'])
        # Expecting 2 more calls - 1 for the port, 1 for the floatingip
        self._verify_event_handler_calls(
            [internal_port, floatingip], expected_call_count=4)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# NOTE(ivar): right now, the flowclassifier plugin doesn't support
# notifications right now. Adding our own using the proper resource name
# could be dangerous for compatibility once they suddenly start supporting
# them. We create our own resource type and make sure to modify it once
# support is added to the SFC project.

from networking_sfc.extensions import flowclassifier

LOGICAL_SRC_NET = 'logical_source_network'
LOGICAL_DST_NET = 'logical_destination_network'
HEALTHCHECK_POLICY = 'healthcheck_policy'
AIM_FLC_L7_PARAMS = {
    LOGICAL_SRC_NET: {
        'allow_post': True, 'allow_put': False,
        'is_visible': True, 'default': None,
        'validate': {'type:uuid_or_none': None}},
    LOGICAL_DST_NET: {
        'allow_post': True, 'allow_put': False,
        'is_visible': True, 'default': None,
        'validate': {'type:uuid_or_none': None}}
}
AIM_PPG_PARAMS = {
    'healthcheck_type': {
        'type:values': ['', 'icmp', 'tcp']
    },
    'healthcheck_frequency': {
        'type:non_negative': None
    },
    'healthcheck_tcp_port': {
        'convert_to': flowclassifier.normalize_port_value
    },
}
AIM_FLC_PARAMS = ['source_ip_prefix', 'destination_ip_prefix']



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class Subnet(object):

    def __init__(self, data):
        self.data = data

    def purge(self):
        if self.data:
            return {
                'cidr': self.data.get('cidr'),
                'id': self.data.get('id'),
                'gateway_ip': self.data.get('gateway_ip'),
                'name': self.data.get('name')
            }
        return self.data


class Port(object):

    def __init__(self, data):
        self.data = data

    def purge(self):
        if self.data:
            return {
                'id': self.data.get('id'),
                'ip_address': self.data.get('ip_address'),
                'mac_address': self.data.get('mac_address'),
                'mac': self.data.get('mac'),
                'name': self.data.get('name'),
                'fixed_ips': self.data.get('fixed_ips'),
                'gateway_ip': self.data.get('gateway_ip'),
                'neutron_port': self.data.get('neutron_port'),
                'cidr': self.data.get('cidr')
            }
        return self.data


class Pt(object):

    def __init__(self, data):
        self.data = data

    def purge(self):
        if self.data:
            return {
                'id': self.data.get('id'),
                'name': self.data.get('name'),
                'port_id': self.data.get('port_id'),
                'policy_target_group_id': self.data.get(
                    'policy_target_group_id'),
                'group_default_gateway': self.data.get(
                    'group_default_gateway'),
                'proxy_gateway': self.data.get(
                    'proxy_gateway')
            }
        return self.data


class Ptg(object):

    def __init__(self, data):
        self.data = data

    def purge(self):
        if self.data:
            return {
                'id': self.data.get('id'),
                'name': self.data.get('name'),
                'provided_policy_rule_sets': self.data.get(
                    'provided_policy_rule_sets'),
                'proxied_group_id': self.data.get(
                    'proxied_group_id'),
                'policy_targets': self.data.get('policy_targets'),
                'tenant_id': self.data.get('tenant_id'),
                'subnets': self.data.get('subnets'),
                'l2_policy_id': self.data.get('l2_policy_id')
            }
        return self.data


class NetworkFunctionDevice(object):

    def __init__(self, data):
        self.data = data

    def purge(self):
        if self.data:
            return {
                'id': self.data.get('id'),
                'interfaces_in_use': self.data.get('interfaces_in_use'),
                'status': self.data.get('status'),
                'mgmt_ip_address': self.data.get('mgmt_ip_address'),
                'monitoring_port_id': self.data.get('monitoring_port_id'),
                'reference_count': self.data.get('reference_count'),
                'mgmt_port_id': self.data.get('mgmt_port_id'),
                'tenant_id': self.data.get('tenant_id'),
            }
        return self.data


class NetworkFunctionInstance(object):

    def __init__(self, data):
        self.data = data

    def purge(self):
        if self.data:
            return {
                'id': self.data.get('id'),
                'status': self.data.get('status'),
                'port_info': self.data.get('port_info'),
                'network_function_device_id': self.data.get(
                    'network_function_device_id'),
                'tenant_id': self.data.get('tenant_id'),
                'name': self.data.get('name')
            }
        return self.data


class NetworkFunction(object):

    def __init__(self, data):
        self.data = data

    def purge(self):
        if self.data:
            return {
                'name': self.data.get('name'),
                'status': self.data.get('status'),
                'service_id': self.data.get('service_id'),
                'config_policy_id': self.data.get('config_policy_id'),
                'service_profile_id': self.data.get('service_profile_id'),
                'service_chain_id': self.data.get('service_chain_id'),
                'id': self.data.get('id'),
                'tenant_id': self.data.get('tenant_id'),
                'network_function_instances': self.data.get(
                    'network_function_instances'),
                'description': self.data.get('description')
            }
        return self.data


class ResourceOwnerContext(object):

    def __init__(self, data):
        self.data = data

    def purge(self):
        if self.data:
            return {
                'admin_token': self.data.get('admin_token'),
                'admin_tenant_id': self.data.get('admin_tenant_id'),
                'tenant_name': self.data.get('tenant_name'),
                'tenant': self.data.get('tenant')
            }
        return self.data


class Management(object):

    def __init__(self, data):
        self.data = data

    def purge(self):
        if self.data:
            return {
                'port': self.data.get('port')
            }
        return self.data


class Provider(object):

    def __init__(self, data):
        self.data = data

    def purge(self):
        if self.data:
            context = {
                'subnet': Subnet(
                    self.data.get('subnet')).purge(),
                'port_model': self.data.get('port_model'),
                'port_classification': self.data.get('port_classification')
            }

            if type(self.data.get('pt')) is list:
                pt_list = []
                for pt in self.data['pt']:
                    pt_list.append(Pt(pt).purge())
                context['pt'] = pt_list
            else:
                context['pt'] = Pt(self.data.get('pt')).purge()

            if type(self.data.get('ptg')) is list:
                ptg_list = []
                for ptg in self.data['ptg']:
                    ptg_list.append(Ptg(ptg).purge())
                context['ptg'] = ptg_list
            else:
                context['ptg'] = Ptg(self.data.get('ptg')).purge()

            if type(self.data.get('port')) is list:
                port_list = []
                for port in self.data['port']:
                    port_list.append(Port(port).purge())
                context['port'] = port_list
            else:
                context['port'] = Port(self.data.get('port')).purge()

            return context
        return self.data


class Consumer(object):

    def __init__(self, data):
        self.data = data

    def purge(self):
        if self.data:
            context = {
                'subnet': Subnet(
                    self.data.get('subnet')).purge(),
                'port_model': self.data.get('port_model'),
                'port_classification': self.data.get('port_classification')
            }
            if type(self.data.get('pt')) is list:
                pt_list = []
                for pt in self.data['pt']:
                    pt_list.append(Pt(pt).purge())
                context['pt'] = pt_list
            else:
                context['pt'] = Pt(self.data.get('pt')).purge()

            if type(self.data.get('ptg')) is list:
                ptg_list = []
                for ptg in self.data['ptg']:
                    ptg_list.append(Ptg(ptg).purge())
                context['ptg'] = ptg_list
            else:
                context['ptg'] = Ptg(self.data.get('ptg')).purge()

            if type(self.data.get('port')) is list:
                port_list = []
                for port in self.data['port']:
                    port_list.append(Port(port).purge())
                context['port'] = port_list
            else:
                context['port'] = Port(self.data.get('port')).purge()

            return context
        return self.data


class ScNodes(object):

    def __init__(self, data):
        self.data = data

    def purge(self):
        if self.data:
            sc_service_profile = self.data.get('sc_service_profile')
            context = {'sc_service_profile': {}}
            if sc_service_profile:
                context['sc_service_profile'][
                    'service_type'] = sc_service_profile.get('service_type')
            return context
        return self.data


class ServiceChainSpecs(object):

    def __init__(self, data):
        self.data = data

    def purge(self):
        if self.data:
            sc_nodes = self.data.get('sc_nodes')
            if type(sc_nodes) is list:
                context = []
                for sc_node in sc_nodes:
                    context.append(ScNodes(sc_node).purge())
                return {
                    'sc_nodes': context
                }
            else:
                return {
                    'sc_nodes': ScNodes(sc_nodes).purge()
                }
        return self.data


class ServiceChainInstance(object):

    def __init__(self, data):
        self.data = data

    def purge(self):
        if self.data:
            return {
                'id': self.data.get('id'),
                'config_param_values': self.data.get('config_param_values'),
                'name': self.data.get('name'),
                'provider_ptg_id': self.data.get('provider_ptg_id')
            }
        return self.data


class ConsumingPtgsDetails(object):

    def __init__(self, data):
        self.data = data

    def purge(self):
        if self.data:
            context = {}
            context['ptg'] = Ptg(self.data.get('ptg')).purge()
            subnets = self.data.get('subnets')
            if type(subnets) is list:
                subnet_ctxt = []
                for subnet in subnets:
                    subnet_ctxt.append(Subnet(subnet).purge())
                context['subnets'] = subnet_ctxt
            else:
                context['subnets'] = Subnet(subnets).purge()
            return context
        return self.data


class ServiceChainNode(object):

    def __init__(self, data):
        self.data = data

    def purge(self):
        if self.data:
            return {
                'service_profile_id': self.data.get('service_profile_id'),
                'service_type': self.data.get('service_type'),
                'config': self.data.get('config'),
                'name': self.data.get('name'),
                'id': self.data.get('id')
            }
        return self.data


class ServiceDetails(object):

    def __init__(self, data):
        self.data = data

    def purge(self):
        if self.data:
            return {
                'service_vendor': self.data.get('service_vendor'),
                'service_type': self.data.get('service_type'),
                'network_mode': self.data.get('network_mode'),
                'image_name': self.data.get('image_name'),
                'device_type': self.data.get('device_type'),
            }
        return self.data


class ConsumingEpsDetails(object):

    def __init__(self, data):
        self.data = data

    def purge(self):
        if self.data:
            return {
                'id': self.data.get('id')
            }
        return self.data


class ServerGrpId(object):

    def __init__(self, data):
        self.data = data

    def purge(self):
        if self.data:
            return {
                'result': self.data.get('result')
            }
        return self.data


class ServiceProfile(object):

    def __init__(self, data):
        self.data = data

    def purge(self):
        if self.data:
            return {
                'id': self.data.get('id'),
                'service_flavor': self.data.get('service_flavor'),
                'service_type': self.data.get('service_type')
            }
        return self.data


class LogContext(object):

    def __init__(self, data):
        self.data = data

    def purge(self):
        if self.data:
            return {
                'meta_id': self.data.get('meta_id', '-'),
                'nfi_id': self.data.get('nfi_id', '-'),
                'nfd_id': self.data.get('nfd_id', '-'),
                'path': self.data.get('path'),
                'auth_token': self.data.get('auth_token'),
                'namespace': self.data.get('namespace')
            }
        return self.data


class NfpContext(object):

    def __init__(self, data):
        self.data = data

    def purge(self):

        context = {
            'active_nfd_ids': self.data.get('active_nfd_ids'),
            'device_without_plugging': self.data.get(
                'device_without_plugging'),
            'id': self.data.get('id'),      # event id
            'key': self.data.get('key'),    # event key
            'admin_token': self.data.get('admin_token'),
            'event_desc': self.data.get('event_desc'),
            'config_policy_id': self.data.get('config_policy_id'),
            'management_ptg_id': self.data.get('management_ptg_id'),
            'network_function_mode': self.data.get('network_function_mode'),
            'files': self.data.get('files'),
            'base_mode_support': self.data.get('base_mode_support'),
            'share_existing_device': self.data.get('share_existing_device'),
            'tenant_id': self.data.get('tenant_id'),
            'binding_key': self.data.get('binding_key'),
            'provider_metadata': self.data.get('provider_metadata'),
            'admin_tenant_id': self.data.get('admin_tenant_id'),
            'is_nfi_in_graph': self.data.get('is_nfi_in_graph'),
            'network_function_device': NetworkFunctionDevice(
                self.data.get('network_function_device')).purge(),
            'network_function_instance': NetworkFunctionInstance(
                self.data.get('network_function_instance')).purge(),
            'network_function': NetworkFunction(
                self.data.get('network_function')).purge(),
            'resource_owner_context': ResourceOwnerContext(
                self.data.get('resource_owner_context')).purge(),
            'management': Management(
                self.data.get('management')).purge(),
            'provider': Provider(
                self.data.get('provider')).purge(),
            'consumer': Consumer(
                self.data.get('consumer')).purge(),
            'service_chain_instance': ServiceChainInstance(
                self.data.get('service_chain_instance')).purge(),
            'service_details': ServiceDetails(
                self.data.get('service_details')).purge(),
            'service_chain_node': ServiceChainNode(
                self.data.get('service_chain_node')).purge(),
            'server_grp_id': ServerGrpId(
                self.data.get('server_grp_id')).purge(),
            'service_profile': ServiceProfile(
                self.data.get('service_profile')).purge(),
            'log_context': LogContext(self.data.get('log_context')).purge(),
            'enable_port_security': self.data.get('enable_port_security')
        }

        service_chain_specs = self.data.get('service_chain_specs')
        if type(service_chain_specs) is list:
            ctxt = []
            for sc_specs in service_chain_specs:
                ctxt.append(ServiceChainSpecs(sc_specs).purge())
            context['service_chain_specs'] = ctxt
        else:
            context['service_chain_specs'] = ServiceChainSpecs(
                service_chain_specs).purge()

        consuming_ptgs_details = self.data.get('consuming_ptgs_details')
        if type(consuming_ptgs_details) is list:
            ctxt = []
            for ptgs_details in consuming_ptgs_details:
                ctxt.append(ConsumingPtgsDetails(ptgs_details).purge())
            context['consuming_ptgs_details'] = ctxt
        else:
            context['consuming_ptgs_details'] = ConsumingPtgsDetails(
                consuming_ptgs_details).purge()

        consuming_eps_details = self.data.get('consuming_eps_details')
        if type(consuming_eps_details) is list:
            ctxt = []
            for eps_details in consuming_eps_details:
                ctxt.append(ConsumingEpsDetails(eps_details).purge())
            context['consuming_eps_details'] = ctxt
        else:
            context['consuming_eps_details'] = ConsumingEpsDetails(
                consuming_eps_details).purge()

        return context
